"""
    QuaO Project device_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.provider_type import ProviderType
from ..enum.sdk import Sdk
from ..model.device.aws_braket_device import AwsBraketDevice
from ..model.device.ibm_cloud_device import IbmCloudDevice
from ..model.device.ibm_quantum_device import IbmQuantumDevice
from ..model.device.quao_braket_device import QuaoBraketDevice
from ..model.device.quao_device import QuaoDevice
from ..model.device.quao_qiskit_device import QuaoQiskitDevice
from ..model.device.quao_tytan_device import QuaoTytanDevice
from ..model.provider.provider import Provider


class DeviceFactory:
    @staticmethod
    def create_device(provider: Provider,
                      device_specification: str,
                      authentication: dict,
                      sdk: Sdk):
        """
        @param sdk:
        @param provider:
        @param device_specification:
        @param authentication:
        @return:
        """

        provider_type = ProviderType.resolve(provider.get_provider_type().value)

        if ProviderType.QUAO_QUANTUM_SIMULATOR.__eq__(provider_type):
            if Sdk.QISKIT.__eq__(sdk):
                return QuaoQiskitDevice(provider, device_specification)

            if Sdk.BRAKET.__eq__(sdk):
                return QuaoBraketDevice(provider, device_specification)

            if Sdk.TYTAN.__eq__(sdk):
                return QuaoTytanDevice(provider, device_specification)

        if ProviderType.IBM_QUANTUM.__eq__(provider_type):
            return IbmQuantumDevice(provider, device_specification)

        if ProviderType.IBM_CLOUD.__eq__(provider_type):
            return IbmCloudDevice(provider, device_specification)

        if ProviderType.AWS_BRAKET.__eq__(provider_type):
            return AwsBraketDevice(provider,
                                   device_specification,
                                   authentication.get('bucketName'),
                                   authentication.get('prefix'))

        raise Exception("Unsupported device!")



"""
    QuaO Project job_fetching.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from typing import Any

from qiskit import QiskitError

from ..callback.update_job_metadata import update_job_metadata
from ...async_tasks.post_processing_task import post_processing_task
from ...config.thread_config import circuit_running_pool
from ...data.callback.callback_url import CallbackUrl
from ...data.promise.post_processing_promise import PostProcessingPromise
from ...data.request.job_fetching_request import JobFetchingRequest
from ...data.response.authentication import Authentication
from ...data.response.job_response import JobResponse
from ...data.response.project_header import ProjectHeader
from ...enum.invocation_step import InvocationStep
from ...config.logging_config import logger
from ...enum.media_type import MediaType
from ...enum.provider_type import ProviderType
from ...enum.sdk import Sdk
from ...enum.status.job_status import JobStatus
from ...enum.status.status_code import StatusCode
from ...factory.provider_factory import ProviderFactory
from ...util.json_parser_utils import JsonParserUtils
from ...util.response_utils import ResponseUtils


class JobFetching:
    def __init__(self, request_data: JobFetchingRequest):
        self.provider_authentication: dict = request_data.provider_authentication
        self.provider_job_id: str = request_data.provider_job_id
        self.backend_authentication: Authentication = request_data.authentication
        self.project_header: ProjectHeader = request_data.project_header
        self.callback_dict: dict = {
            InvocationStep.ANALYSIS: request_data.analysis,
            InvocationStep.FINALIZATION: request_data.finalization
        }

    def fetch(self, post_processing_fn):
        """

        @param post_processing_fn:
        @return:
        """
        logger.debug("[Fetching] Fetch job")

        job_response = JobResponse(
            provider_job_id=self.provider_job_id,
            authentication=self.backend_authentication,
            project_header=self.project_header,
            status_code=StatusCode.DONE)

        try:
            provider = ProviderFactory \
                .create_provider(provider_type=ProviderType.IBM_QUANTUM,
                                 sdk=Sdk.QISKIT,
                                 authentication=self.provider_authentication) \
                .collect_provider()

            job = provider.retrieve_job(job_id=self.provider_job_id)

            job_response.job_status = job.status().name

            if JobStatus.DONE.value.__eq__(job_response.job_status):
                circuit_running_pool.submit(self.__handle_job_result,
                                            job.result(),
                                            job_response,
                                            self.callback_dict,
                                            post_processing_fn)
            elif JobStatus.ERROR.value.__eq__(job_response.job_status):
                job_response.job_result = JsonParserUtils.parse(job.result())
            else:
                job_response.status_code = StatusCode.POLLING

        except Exception as exception:
            logger.debug("Exception when fetch job with provider_job_id {0}: {1}".format(
                self.provider_job_id, str(exception)))

            job_response.job_result = {
                "error": "Exception when fetch job with provider_job_id {0}: {1}".format(
                    self.provider_job_id, str(exception)),
                "exception": str(exception),
            }
            job_response.status_code = StatusCode.ERROR
            job_response.job_status = JobStatus.ERROR.value

        return ResponseUtils.generate_response(job_response)

    def __handle_job_result(self,
                            original_job_result,
                            job_response: JobResponse,
                            callback_dict: dict,
                            post_processing_fn):
        """
        Fetch job from IBM Quantum

        @return: Job status
        """

        job_response = self.__on_analysis(
            callback_url=callback_dict.get(InvocationStep.ANALYSIS),
            job_response=job_response,
            original_job_result=original_job_result)

        if job_response is None:
            return

        self.__on_finalization(post_processing_fn=post_processing_fn,
                               callback_url=callback_dict.get(InvocationStep.FINALIZATION),
                               original_job_result=original_job_result)

    def __on_analysis(self,
                      callback_url: CallbackUrl,
                      job_response: JobResponse,
                      original_job_result):
        """

        @param callback_url:
        @param job_response:
        @param original_job_result:
        @return:
        """
        logger.debug("[Fetching] On analysis")

        update_job_metadata(job_response=job_response,
                            callback_url=callback_url.on_start)

        try:
            job_response.content_type = MediaType.APPLICATION_JSON

            job_response.job_histogram = self.__produce_histogram_data(original_job_result)

            job_result_parse = JsonParserUtils.parse(original_job_result)

            job_response.execution_time = self.__get_execution_time(job_result_parse)

            update_job_metadata(
                job_response=job_response,
                callback_url=callback_url.on_done)

            return job_response

        except Exception as exception:
            logger.error("Exception when analyst job result with provider_job_id {0}: {1}".format(
                self.provider_job_id, str(exception)))

            job_response.status_code = StatusCode.ERROR
            job_response.job_status = JobStatus.ERROR.value
            job_response.job_result = {
                "error": "Exception when analyst job result with provider_job_id {0}".format(
                    self.provider_job_id),
                "exception": str(exception),
            }

            update_job_metadata(
                job_response=job_response,
                callback_url=callback_url.on_error)

            return None

    def __on_finalization(self,
                          post_processing_fn,
                          callback_url: CallbackUrl,
                          original_job_result):
        """

        @param post_processing_fn:
        @param callback_url:
        @param original_job_result:
        """
        logger.debug("[Fetching] On finalization")

        promise = PostProcessingPromise(callback_url=callback_url,
                                        authentication=self.backend_authentication,
                                        job_result=original_job_result,
                                        project_header=self.project_header)

        post_processing_task(post_processing_fn=post_processing_fn,
                             promise=promise)

    @staticmethod
    def __produce_histogram_data(job_result) -> Any | None:
        """

        @param job_result:
        @return:
        """
        logger.debug("[Fetching] Produce histogram")

        try:
            return job_result.get_counts()
        except QiskitError as qiskit_error:
            logger.debug("Can't produce histogram with error: {0}".format(str(qiskit_error)))
            return None

    @staticmethod
    def __get_execution_time(job_result):
        """

        @param job_result:
        @return:
        """
        logger.debug("[Fetching] Get execution time")

        if "metadata" not in job_result:
            return None

        metadata = job_result["metadata"]

        if (
                metadata is None
                or not bool(metadata)
                or "time_taken_execute" not in metadata
        ):
            return None

        return metadata["time_taken_execute"]



"""
    QuaO Project quao_qiskit_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_aer import Aer

from ...config.logging_config import logger
from ...model.provider.quao_provider import QuaoProvider


class QuaoQiskitProvider(QuaoProvider):
    def get_backend(self, device_specification):
        logger.debug('[Quao Qiskit] Get backend')

        provider = self.collect_provider()

        device_names = set(map(self.__map_aer_backend_name, provider.backends()))

        if device_names.__contains__(device_specification):
            backend = provider[0].get_backend(device_specification)

            return backend

        raise Exception('Unsupported device')

    def collect_provider(self):
        logger.debug('[Quao Qiskit] Connect to provider')

        return Aer

    @staticmethod
    def __map_aer_backend_name(backend):
        return backend.configuration().backend_name



"""
    QuaO Project post_processing_task.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..component.callback.update_job_metadata import update_job_metadata
from ..config.logging_config import logger
from ..data.promise.post_processing_promise import PostProcessingPromise
from ..data.response.job_response import JobResponse
from ..enum.media_type import MediaType
from ..enum.status.job_status import JobStatus
from ..enum.status.status_code import StatusCode
from ..util.json_parser_utils import JsonParserUtils


def post_processing_task(post_processing_fn, promise: PostProcessingPromise):
    """
    Execute post_processing and send to backend

    @param post_processing_fn: post_processing function
    @param promise: promise arg
    """

    job_response = JobResponse(
        authentication=promise.authentication,
        project_header=promise.project_header,
        job_status=JobStatus.DONE.value,
        status_code=StatusCode.DONE)

    update_job_metadata(job_response=job_response,
                        callback_url=promise.callback_url.on_start)

    job_response.content_type = MediaType.APPLICATION_JSON

    try:

        logger.info("Execute post_processing ...")
        job_result_post_processing = post_processing_fn(promise.job_result)
        logger.info("Execute post_processing finishing!")

        logger.debug("Parsing job result....")
        job_response.job_result = JsonParserUtils.parse(job_result_post_processing)
        logger.debug("Parsing job result completed!")

        update_job_metadata(job_response=job_response,
                            callback_url=promise.callback_url.on_done)

    except Exception as exception:
        job_response.job_result = {
            "error": "Error when execute post_processing(): ",
            "exception": str(exception),
        }
        job_response.status_code = StatusCode.ERROR
        job_response.job_status = JobStatus.ERROR.value

        update_job_metadata(job_response=job_response,
                            callback_url=promise.callback_url.on_error)



"""
    QuaO Project device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import abstractmethod, ABC

from ...async_tasks.post_processing_task import post_processing_task
from ...component.callback.update_job_metadata import update_job_metadata
from ...config.logging_config import logger
from ...data.callback.callback_url import CallbackUrl
from ...data.device.circuit_running_option import CircuitRunningOption
from ...data.promise.post_processing_promise import PostProcessingPromise
from ...data.response.authentication import Authentication
from ...data.response.job_response import JobResponse
from ...data.response.project_header import ProjectHeader
from ...enum.invocation_step import InvocationStep
from ...enum.media_type import MediaType
from ...enum.status.job_status import JobStatus
from ...enum.status.status_code import StatusCode
from ...model.provider.provider import Provider
from ...util.json_parser_utils import JsonParserUtils


class Device(ABC):
    def __init__(self, provider: Provider, device_specification: str):
        self.provider = provider
        self.device = provider.get_backend(device_specification)
        self.execution_time = None

    def run_circuit(self,
                    circuit,
                    post_processing_fn,
                    options: CircuitRunningOption,
                    callback_dict: dict,
                    authentication: Authentication,
                    project_header: ProjectHeader):
        """

        @param project_header: project header
        @param callback_dict: callback url dictionary
        @param options: Options for run circuit
        @param authentication: Authentication for calling quao server
        @param post_processing_fn: Post-processing function
        @param circuit: Circuit was run
        """

        original_job_result, job_response = self.__on_execution(
            authentication=authentication,
            project_header=project_header,
            execution_callback=callback_dict.get(InvocationStep.EXECUTION),
            circuit=circuit,
            options=options)

        if original_job_result is None:
            return

        job_response = self.__on_analysis(
            job_response=job_response,
            original_job_result=original_job_result,
            analysis_callback=callback_dict.get(InvocationStep.ANALYSIS))

        if job_response is None:
            return

        self.__on_finalization(job_result=original_job_result,
                               authentication=authentication,
                               post_processing_fn=post_processing_fn,
                               finalization_callback=callback_dict.get(InvocationStep.FINALIZATION),
                               project_header=project_header)

    def __on_execution(self, authentication: Authentication,
                       project_header: ProjectHeader,
                       execution_callback: CallbackUrl,
                       circuit,
                       options: CircuitRunningOption):
        """

        @param authentication: authentication information
        @param project_header: project header information
        @param execution_callback: execution step callback urls
        @param circuit: circuit will be run
        @param options: options will use for running
        @return: job and job response
        """
        logger.debug("[Invocation] On execution")

        job_response = JobResponse(authentication=authentication,
                                   project_header=project_header,
                                   status_code=StatusCode.DONE)

        update_job_metadata(job_response=job_response,
                            callback_url=execution_callback.on_start)
        try:
            job = self._create_job(circuit=circuit, options=options)
            job_response.provider_job_id = self._get_provider_job_id(job)
            job_response.job_status = self._get_job_status(job)
            original_job_result = None

            if self._is_simulator():
                job_result = self._get_job_result(job)
                job_response.job_status = self._get_job_status(job)

                if JobStatus.ERROR.value.__eq__(job_response.job_status):
                    job_response.job_result = job_result
                    job_response.content_type = MediaType.APPLICATION_JSON
                else:
                    original_job_result = job_result
            else:
                job_response.status_code = StatusCode.POLLING

            update_job_metadata(job_response=job_response,
                                callback_url=execution_callback.on_done)

            return original_job_result, job_response

        except Exception as exception:
            logger.debug("Execute job failed with error {}".format(str(exception)))

            job_response.status_code = StatusCode.ERROR
            job_response.content_type = MediaType.APPLICATION_JSON
            job_response.job_status = JobStatus.ERROR.value
            job_response.job_result = {"error": str(exception)}

            update_job_metadata(job_response=job_response,
                                callback_url=execution_callback.on_error)
            return None, None

    def __on_analysis(self, job_response: JobResponse,
                      analysis_callback: CallbackUrl,
                      original_job_result):
        """

        @param job_response:
        @param analysis_callback:
        @param original_job_result:
        @return:
        """
        logger.debug("[Invocation] On analysis")

        update_job_metadata(job_response=job_response,
                            callback_url=analysis_callback.on_start)

        try:
            job_response.job_histogram = self._produce_histogram_data(original_job_result)

            original_job_result_dict = JsonParserUtils.parse(original_job_result)
            self._calculate_execution_time(original_job_result_dict)
            job_response.execution_time = self.execution_time

            update_job_metadata(
                job_response=job_response,
                callback_url=analysis_callback.on_done)

            return job_response

        except Exception as exception:
            logger.error("Invocation - Exception when analyst job result : {0}".format(
                str(exception)))

            job_response.status_code = StatusCode.ERROR
            job_response.content_type = MediaType.APPLICATION_JSON
            job_response.job_status = JobStatus.ERROR.value
            job_response.job_result = {"error": str(exception)}

            update_job_metadata(job_response=job_response,
                                callback_url=analysis_callback.on_error)
            return None

    @staticmethod
    def __on_finalization(job_result,
                          finalization_callback: CallbackUrl,
                          post_processing_fn,
                          authentication: Authentication,
                          project_header: ProjectHeader):
        """

        @param job_result: final job result
        @param finalization_callback: callback for update job result
        @param post_processing_fn: post processing function
        @param authentication: authentication of quao
        @param project_header: project header
        """

        if finalization_callback is None:
            return

        logger.info("[Invocation] On finalization")

        post_processing_promise = PostProcessingPromise(
            callback_url=finalization_callback,
            authentication=authentication,
            job_result=job_result,
            project_header=project_header)

        post_processing_task(post_processing_fn,
                             post_processing_promise)

    @abstractmethod
    def _create_job(self, circuit, options: CircuitRunningOption):
        """

        @param circuit: Circuit for create job
        @param options:
        """

        raise NotImplemented('[Device] _create_job() method must be implemented')

    @abstractmethod
    def _is_simulator(self) -> bool:
        """

        """

        raise NotImplemented('[Device] _is_simulator() method must be implemented')

    @abstractmethod
    def _produce_histogram_data(self, job_result) -> dict | None:
        """

        @param job_result:
        """

        raise NotImplemented('[Device] _produce_histogram_data() method must be implemented')

    @abstractmethod
    def _get_provider_job_id(self, job) -> str:
        """

        """

        raise NotImplemented('[Device] _get_provider_job_id() method must be implemented')

    @abstractmethod
    def _get_job_status(self, job) -> str:
        """

        """

        raise NotImplemented('[Device] _get_job_status() method must be implemented')

    @abstractmethod
    def _calculate_execution_time(self, job_result) -> float:
        """

        """

        raise NotImplemented('[Device] _calculate_execution_time() method must be implemented')

    def _get_job_result(self, job):
        """

        @param job:
        @return:
        """
        logger.debug('[Device] Get job result')

        return job.result()



"""
    QuaO Project backend.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from braket.circuits import Circuit
from qiskit import QuantumCircuit

from ..callback.update_job_metadata import update_job_metadata
from ...async_tasks.export_circuit_task import export_circuit_task
from ...component.device.device_selection import DeviceSelection
from ...config.logging_config import logger
from ...config.thread_config import circuit_exporting_pool
from ...config.thread_config import circuit_running_pool
from ...data.backend.backend_information import BackendInformation
from ...data.device.circuit_running_option import CircuitRunningOption
from ...data.request.invocation_request import InvocationRequest
from ...data.response.authentication import Authentication
from ...data.response.job_response import JobResponse
from ...data.response.project_header import ProjectHeader
from ...enum.invocation_step import InvocationStep
from ...enum.media_type import MediaType
from ...enum.provider_type import ProviderType
from ...enum.sdk import Sdk
from ...enum.status.job_status import JobStatus
from ...enum.status.status_code import StatusCode
from ...factory.device_factory import DeviceFactory
from ...factory.provider_factory import ProviderFactory
from ...util.circuit_utils import CircuitUtils


class Invocation:
    def __init__(self, request_data: InvocationRequest):
        self.sdk: Sdk = Sdk.resolve(request_data.sdk)
        self.input = request_data.input
        self.device_id = request_data.device_id
        self.backend_information: BackendInformation
        self.authentication: Authentication = request_data.authentication
        self.project_header: ProjectHeader = request_data.project_header
        self.options = CircuitRunningOption(shots=request_data.shots,
                                            processing_unit=request_data.processing_unit,
                                            executor=circuit_running_pool,
                                            max_job_size=1)
        self.device_selection_url: str = request_data.device_selection_url
        self.circuit_export_url: str = request_data.circuit_export_url
        self.callback_dict: dict = {
            InvocationStep.PREPARATION: request_data.preparation,
            InvocationStep.EXECUTION: request_data.execution,
            InvocationStep.ANALYSIS: request_data.analysis,
            InvocationStep.FINALIZATION: request_data.finalization
        }

    def submit_job(self, circuit_preparation_fn, post_processing_fn):
        """

        @param post_processing_fn: Post-processing function
        @param circuit_preparation_fn: Circuit-preparation function
        @return: Job result
        """
        logger.debug("[Invocation] Invoke job")

        circuit = self.__pre_execute(circuit_preparation_fn)

        if circuit is None:
            return

        self.__execute(circuit, post_processing_fn)

    def __pre_execute(self, circuit_preparation_fn):
        """

        @param circuit_preparation_fn: Circuit preparation function
        """
        circuit = self.__prepare_circuit(circuit_preparation_fn)

        if circuit is None:
            return None

        self.__prepare_backend_data(circuit)

        if not Sdk.TYTAN.__eq__(self.sdk):
            self.__export_circuit(circuit)

        return circuit

    def __execute(self, circuit, post_processing_fn):
        """

        @param circuit: Circuit was run
        @param post_processing_fn: Post-processing function
        @return: Job response
        """

        logger.debug('[Invocation] Execute job')

        try:
            if self.backend_information is None:
                raise Exception("Backend is not found")

            device_name = self.backend_information.device_name
            provider_tag = self.backend_information.provider_tag
            backend_authentication = self.backend_information.authentication

            logger.debug('[Invocation] Execute job with provider tag: {0}'.format(provider_tag))
            provider = ProviderFactory.create_provider(
                provider_type=ProviderType.resolve(provider_tag),
                sdk=self.sdk,
                authentication=backend_authentication)

            logger.debug('[Invocation] Execute job with device name: {0}'.format(device_name))
            device = DeviceFactory.create_device(provider=provider,
                                                 device_specification=device_name,
                                                 authentication=backend_authentication,
                                                 sdk=self.sdk)

        except Exception as exception:
            job_response = JobResponse(job_status=JobStatus.ERROR.value,
                                       content_type=MediaType.APPLICATION_JSON,
                                       status_code=StatusCode.ERROR,
                                       authentication=self.authentication,
                                       project_header=self.project_header,
                                       job_result={"error": str(exception)})
            update_job_metadata(
                job_response=job_response,
                callback_url=self.callback_dict.get(InvocationStep.EXECUTION).on_error)

            return

        device.run_circuit(circuit=circuit,
                           post_processing_fn=post_processing_fn,
                           options=self.options,
                           callback_dict=self.callback_dict,
                           authentication=self.authentication,
                           project_header=self.project_header)

    def __prepare_circuit(self, circuit_preparation_fn):
        """

        @param circuit_preparation_fn: Circuit preparation function
        @return: circuit
        """
        logger.debug("[Invocation] Prepare circuit")

        job_response = JobResponse(status_code=StatusCode.DONE,
                                   authentication=self.authentication,
                                   project_header=self.project_header)
        update_job_metadata(
            job_response=job_response,
            callback_url=self.callback_dict.get(InvocationStep.PREPARATION).on_start)

        try:
            circuit = circuit_preparation_fn(self.input)

            if circuit is None:
                raise Exception("Invalid circuit")

            update_job_metadata(
                job_response=job_response,
                callback_url=self.callback_dict.get(InvocationStep.PREPARATION).on_done)

            return circuit

        except Exception as exception:
            job_response.job_result = JobStatus.ERROR.value
            job_response.content_type = MediaType.APPLICATION_JSON
            job_response.status_code = StatusCode.ERROR
            job_response.job_result = {"error": str(exception)}

            update_job_metadata(
                job_response=job_response,
                callback_url=self.callback_dict.get(InvocationStep.PREPARATION).on_error)

            return None

    def __prepare_backend_data(self, circuit):
        """

        @param circuit: Circuit was run
        """

        required_qubit_amount = CircuitUtils.get_qubit_amount(circuit)

        device_selection = DeviceSelection(required_qubit_amount,
                                           self.device_id,
                                           self.authentication.user_token,
                                           self.device_selection_url)

        self.backend_information = device_selection.select()

    def __export_circuit(self, circuit):
        """

        @param circuit: Circuit was exported
        """

        circuit_config_map = {
            'circuit': circuit,
            'circuit_export_url': self.circuit_export_url
        }

        backend_config_map = {
            'provider_tag': self.backend_information.provider_tag,
            'device_name': self.backend_information.device_name,
            'authentication': self.backend_information.authentication
        }

        circuit_exporting_pool.submit(export_circuit_task,
                                      circuit_config_map,
                                      backend_config_map,
                                      self.authentication.user_token)



"""
    QuaO Project backend_information.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""


class BackendInformation:
    def __init__(self,
                 device_name: str,
                 provider_tag: str,
                 authentication: dict):
        self.device_name = device_name
        self.provider_tag = provider_tag
        self.authentication = authentication



"""
    QuaO Project device_selection.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from json import JSONDecodeError

import requests

from ...config.logging_config import logger
from ...data.backend.backend_information import BackendInformation
from ...util.http_utils import HttpUtils


class DeviceSelection:
    def __init__(self,
                 required_qubit_number,
                 device_id,
                 user_token,
                 url):
        self.required_qubit_number = required_qubit_number
        self.device_id = device_id
        self.user_token = user_token
        self.url = url

    def select(self) -> BackendInformation:
        """

        @return: Backend information
        """

        request, header = self._prepare()

        return self._do(request, header)

    def _prepare(self):
        """

        @return: Request and header
        """

        request = {
            "deviceId": self.device_id,
            "qubitAmount": self.required_qubit_number
        }
        header = HttpUtils.create_bearer_header(self.user_token)

        return request, header

    def _do(self, request: dict, header: dict):
        """

        @param request: Request
        @param header: Header
        @return:
        """
        logger.debug('Select device with request: {0}'.format(request))

        response = requests.get(
            self.url,
            params=request,
            headers=header
        )

        if response.status_code != 200:
            logger.debug('Select device fail with status code: {0}'.format(response.status_code))

            return None

        try:
            response_dict = response.json().get("data")

            logger.debug('Select device successfully: {0}'.format(response_dict))

            return BackendInformation(
                provider_tag=response_dict.get("providerTag"),
                device_name=response_dict.get("deviceName"),
                authentication=response_dict.get("authentication"))
        except JSONDecodeError:
            return None



"""
    QuaO Project provider_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.provider_type import ProviderType
from ..enum.sdk import Sdk
from ..model.provider.aws_braket_provider import AwsBraketProvider
from ..model.provider.ibm_cloud_provider import IbmCloudProvider
from ..model.provider.ibm_quantum_provider import IbmQuantumProvider
from ..model.provider.quao_braket_provider import QuaoBraketProvider
from ..model.provider.quao_qiskit_provider import QuaoQiskitProvider
from ..model.provider.quao_tytan_provider import QuaoTytanProvider


class ProviderFactory:
    @staticmethod
    def create_provider(provider_type: ProviderType,
                        sdk: Sdk,
                        authentication: dict):
        """

        @param sdk:
        @param provider_type:
        @param authentication:
        @return:
        """

        if ProviderType.QUAO_QUANTUM_SIMULATOR.__eq__(provider_type):
            if Sdk.QISKIT.__eq__(sdk):
                return QuaoQiskitProvider()

            if Sdk.BRAKET.__eq__(sdk):
                return QuaoBraketProvider()

            if Sdk.TYTAN.__eq__(sdk):
                return QuaoTytanProvider()

        if ProviderType.IBM_QUANTUM.__eq__(provider_type):
            return IbmQuantumProvider(authentication.get("token"))

        if ProviderType.IBM_CLOUD.__eq__(provider_type):
            return IbmCloudProvider(authentication.get("token"),
                                    authentication.get("crn"))

        if ProviderType.AWS_BRAKET.__eq__(provider_type):
            return AwsBraketProvider(authentication.get("accessKey"),
                                     authentication.get("secretKey"),
                                     authentication.get("regionName"))

        raise Exception("Unsupported provider!")



"""
    QuaO Project job_status.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from enum import Enum


class JobStatus(Enum):
    ERROR = 'ERROR'
    COMPLETED = 'COMPLETED'
    DONE = 'DONE'



"""
    QuaO Project callback_url.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""


class CallbackUrl:
    def __init__(self, callback_url: dict):
        self.on_start = callback_url.get("onStartCallbackUrl")
        self.on_error = callback_url.get("onErrorCallbackUrl")
        self.on_done = callback_url.get("onDoneCallbackUrl")



"""
    QuaO Project job_response.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from .authentication import Authentication
from .project_header import ProjectHeader
from ...enum.media_type import MediaType
from ...enum.status.status_code import StatusCode


class JobResponse(object):
    def __init__(
            self,
            provider_job_id: str = "",
            job_status: str = None,
            status_code: StatusCode = None,
            job_result: dict = None,
            content_type: MediaType = MediaType.ALL_TYPE,
            authentication: Authentication = None,
            project_header: ProjectHeader = None,
            job_histogram: dict = None,
            execution_time: float = None
    ):
        self.provider_job_id = provider_job_id
        self.job_status = job_status
        self.job_result = job_result
        self.content_type = content_type
        self.job_histogram = job_histogram
        self.user_identity = authentication.user_identity
        self.user_token = authentication.user_token
        self.execution_time = execution_time
        self.status_code = status_code
        self.project_header = project_header





"""
    QuaO Project http_header.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.base_enum import BaseEnum


class HttpHeader(BaseEnum):
    AUTHORIZATION = 'Authorization'
    CONTENT_TYPE = 'Content-type'
    ACCEPT = 'Accept'




"""
    QuaO Project http_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..data.response.project_header import ProjectHeader
from ..enum.http_header import HttpHeader
from ..enum.media_type import MediaType
from ..enum.token_type import TokenType


class HttpUtils:
    @staticmethod
    def create_bearer_header(token):
        """

        @param token:
        @return:
        """

        return {
            HttpHeader.AUTHORIZATION.value: TokenType.BEARER.value + ' ' + token
        }

    @staticmethod
    def create_application_json_header(token: str, project_header: ProjectHeader):
        """

        @param project_header:
        @param token:
        @return:
        """

        return {
            HttpHeader.AUTHORIZATION.value: TokenType.BEARER.value + ' ' + token,
            HttpHeader.CONTENT_TYPE.value: MediaType.APPLICATION_JSON.value,
            HttpHeader.ACCEPT.value: MediaType.APPLICATION_JSON.value,
            project_header.name: project_header.value
        }



"""
    QuaO Project quao_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import ABC

from ...enum.provider_type import ProviderType
from ...model.provider.provider import Provider


class QuaoProvider(Provider, ABC):
    def __init__(self):
        super().__init__(ProviderType.QUAO_QUANTUM_SIMULATOR)



"""
    QuaO Project job_fetching_request.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from .request import Request


class JobFetchingRequest(Request):
    def __init__(self, request_data):
        super().__init__(request_data)
        self.provider_authentication = request_data.get("providerAuthentication")



"""
    QuaO Project handler_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..handler.handler import Handler
from ..handler.invocation_handler import InvocationHandler
from ..handler.job_fetching_handler import JobFetchingHandler


class HandlerFactory:
    @staticmethod
    def create_handler(event,
                       circuit_preparation_fn,
                       post_processing_fn, ) -> Handler:
        request_data = event.json()
        provider_job_id = request_data.get("providerJobId")

        if provider_job_id is None:
            return InvocationHandler(request_data=request_data,
                                     circuit_preparation_fn=circuit_preparation_fn,
                                     post_processing_fn=post_processing_fn)

        return JobFetchingHandler(request_data=request_data,
                                  post_processing_fn=post_processing_fn)



"""
    QuaO Project invocation_step.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.base_enum import BaseEnum


class InvocationStep(BaseEnum):
    POLLING = "POLLING"
    PREPARATION = "PREPARATION"
    EXECUTION = "EXECUTION"
    ANALYSIS = "ANALYSIS"
    FINALIZATION = "FINALIZATION"
    PROMISE = "PROMISE"



"""
    QuaO Project quao_tytan_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from tytan.sampler import SASampler, GASampler

from ...config.logging_config import logger
from ...model.provider.quao_provider import QuaoProvider


class QuaoTytanProvider(QuaoProvider):

    def get_backend(self, device_specification):
        logger.debug('[Quao Tytan] Get backend')

        backend = self.collect_provider().get(device_specification)

        if backend is None:
            raise Exception('Unsupported device')

        return backend

    def collect_provider(self):
        logger.debug('[Quao Tytan] Connect to provider')

        return {
            'SASampler': SASampler(),
            'GASampler': GASampler()
        }



"""
    QuaO Project quao_qiskit_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from qiskit import transpile, QiskitError

from ...config.logging_config import logger
from ...data.device.circuit_running_option import CircuitRunningOption
from ...model.device.quao_device import QuaoDevice


class QuaoQiskitDevice(QuaoDevice):
    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug('[Quao Qiskit] Create job with {0} shots'.format(options.shots))

        self.device.set_options(device=options.processing_unit.value,
                                shots=options.shots,
                                executor=options.executor,
                                max_job_size=options.max_job_size)
        transpiled_circuit = transpile(circuits=circuit, backend=self.device)

        return self.device.run(transpiled_circuit)

    def _produce_histogram_data(self, job_result) -> dict:
        logger.debug('[Quao Qiskit] Produce histogram')

        try:
            histogram_data = job_result.get_counts()
        except QiskitError as qiskit_error:
            logger.debug("Can't produce histogram with error: {0}".format(str(qiskit_error)))
            histogram_data = None

        return histogram_data

    def _get_provider_job_id(self, job) -> str:
        logger.debug('[Quao Qiskit] Get provider job id')

        return job.job_id()

    def _get_job_status(self, job) -> str:
        logger.debug('[Quao Qiskit] Get job status')

        return job.status().name

    def _calculate_execution_time(self, job_result):
        logger.debug('[Quao Qiskit] calculate execution time')

        if "metadata" not in job_result:
            return None

        metadata = job_result["metadata"]

        if (
                metadata is None
                or not bool(metadata)
                or "time_taken_execute" not in metadata
        ):
            return None

        self.execution_time = metadata["time_taken_execute"]

        logger.debug('[Quao Qiskit] Execution time calculation was: {0} seconds'
                     .format(self.execution_time))



"""
    QuaO Project export_circuit.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

import io

import requests
from braket.circuits import Circuit
from qbraid import circuit_wrapper
from qiskit import transpile, Aer

from ..config.logging_config import logger
from ..enum.media_type import MediaType
from ..enum.provider_type import ProviderType
from ..enum.sdk import Sdk
from ..factory.provider_factory import ProviderFactory
from ..util.http_utils import HttpUtils


def export_circuit_task(
        circuit_config_map: dict,
        backend_config_map: dict,
        user_token: str):
    """
      Export circuit to svg file then send to QuaO server for saving
      Args:
          circuit: circuit will be exported
          @param circuit_config_map: Circuit config map
          @param backend_config_map: Backend config map
          @param user_token: User token
    """
    logger.debug("[Circuit export] Start")

    circuit_export_url = circuit_config_map.get('circuit_export_url')
    if circuit_export_url is None or len(circuit_export_url) < 1:
        return

    logger.debug("[Circuit export] Preparing circuit figure...")
    transpiled_circuit = transpile_circuit(circuit_config_map.get('circuit'), backend_config_map)
    circuit_figure = transpiled_circuit.draw(output='mpl', fold=-1)

    logger.debug("[Circuit export] Converting circuit figure to svg file...")
    figure_buffer = io.BytesIO()
    circuit_figure.savefig(figure_buffer, format='svg', bbox_inches='tight')

    logger.debug("[Circuit export] Sending circuit svg image to [{0}] with POST method ...".format(
        circuit_export_url))

    payload = {'circuit': (
        'circuit_figure',
        figure_buffer.getvalue(),
        MediaType.MULTIPART_FORM_DATA.value)}

    response = requests.post(url=circuit_export_url,
                             headers=HttpUtils.create_bearer_header(user_token),
                             files=payload)

    if response.ok:
        logger.debug("Sending request to QuaO backend successfully!")
    else:
        logger.debug("Sending request to QuaO backend failed with status {0}!".format(
            response.status_code))

    logger.debug("[Circuit export] Finish")


def transpile_circuit(circuit, backend_config_map: dict):
    """

    @param circuit: Circuit will be transpiled
    @param backend_config_map: Backend config map
    @return: Transpiled circuit
    """

    if isinstance(circuit, Circuit):
        return circuit_wrapper(circuit).transpile(Sdk.QISKIT.value)

    provider_type = backend_config_map.get('provider_tag')

    if not ProviderType.QUAO_QUANTUM_SIMULATOR.value.__eq__(provider_type):
        provider_type = ProviderType.IBM_QUANTUM.value

    provider = ProviderFactory.create_provider(
        provider_type=provider_type,
        authentication=backend_config_map.get('authentication'))

    backend = provider.get_backend(backend_config_map.get('device_name'))

    return transpile(circuits=circuit, backend=backend)



"""
    QuaO Project quao_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import ABC

from ..device.device import Device
from ..provider.provider import Provider
from ...config.logging_config import logger


class QuaoDevice(Device, ABC):

    def __init__(self, provider: Provider,
                 device_specification: str):
        super().__init__(provider, device_specification)

    def _is_simulator(self) -> bool:
        logger.debug('[Quao] Get device type')

        return True

    def _calculate_execution_time(self, job_result):
        logger.debug('[Quao] Execution time calculation was: {0} seconds'
                     .format(self.execution_time))

    def _get_job_result(self, job):
        logger.debug('[Quao] Get job result')

        return job.result()
